/*
 * Midtomup
 * 
 * Convert Midi files to MUP input files
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include "midifile.h"
#include "mtm.h"

#define TIE 1
#define NO_TIE 0
#define MEASURE -11
#define MAXVEL   500

// global variables - see extern definitions in mtm.h for comments
long bytecount;		// position in .midi file
unsigned char debugbuff[32];
Bstyle bstyles[50];
int tied[MAXTRK];
int key[MAXTRK];
int hkey[MAXTRK];
short tr[MAXTRK];
short vel[MAXTRK];		/* note of last onvelocity used */
int accnotes[TABSIZE];
int sccnotes[TABSIZE][3];
track tracks[MAXTRK];	/* main event tracks */
track mtrks[MAXTRK];	/* secondary tracks for modifiers */
Tsig tsig[100];
Key midikey[MAXKEYSIG];
int sccwhre;
int accwhre;
int triplet;
int velfudge;
int shortest;
int bwhre;
int mwhre;		/* number of midi key changes */
short measure;
short twhre;
long ticks, endticks, h_ticks, lasttick;
track *ct, *mct;		/* current track, and modifiers */
FILE *out;

// local to this file
typedef struct {
  int trk;		// midi track number
  char *map1;		// mup staff + voice
  char *map2;		// staff only (where needed)
} Tlate;

static Tlate tlate[MAXTRK];
static int tlwhre;

static FILE *inf;
static int prevlen = -1;
static int division;        /* from the file header */
// static long tempo = 500000; /* the default tempo is 120 beats/minute */

static int ffactor = 0;	/* default quantization */
static int trackno = 0;
static int zflag;	/* -z flag :=: discard very short notes */
static int cflag;	/* -c flag :=: suppress comment */
static int suppMidi= 1;	/* -m flag :=: suppress midi statement in output */
static int stretch = 0;	/* -s flag :=: amount (as percentage) to be added to each length */
static event *secondaries;
static int modeval;
#define SCORE 1
#define MUSIC 2
#define STAFF 3

static int debug = 0;
static int warn = 1;
static int dfltkey=0;
static int pmeasure = 1;
static int velct;
static int donestaffs = 0;
static int format;		// Midifile format 0,1[,2?]
static long discards = 0;	// notes discarded (too short)
static long velt[MAXVEL];
static int velvel[MAXVEL];
static char ifile[80];
static event *reuse;
static int maxkeysig, maxtimesig;
static char *nnam[] = { "C", "C#", "D", "D#", "E", "F", "F#", "G",
			"G#", "A", "A#", "B" };
#ifndef VERSION
#define VERSION "Version unknown"
#endif
static char *version = "Version "VERSION;

extern char *optarg;
extern int optind;

static event *alloc_event(void) {
  event *e = reuse;

  if (e == NULL)
    e = (event *)malloc(sizeof(event));
  else
    reuse = e->next;
  e->next = NULL;
  e->prev = NULL;
  return e;
}

static void free_event(event *e) {
  e->next = reuse;
  //  reuse = e;
}

static void extractTsig(char *p) {
  int d=0, n=0;

  p +=5;
  while (*p == ' ' || *p == '\t')
    p++;
  while (isdigit(*p))
    n = 10 * n + *p++ - '0';
  if (*p != '/')
    fprintf(out, "// invalid time signature in #header\n");
  else
    p++;
  while (isdigit(*p))
    d = 10 * d + *p++ - '0';
  if (n > 0 && d > 0) {
    tsig[0].num  = n;
    tsig[0].dnom = d;
    twhre = 1;
  }
}

static void doheader(FILE *fp) {
  char buf[80];
  char *p;

  while(fgets(buf, 80, fp) != '\0') {
    if (strncmp("#headend", buf, 8) == 0 ||
	strncmp("#endhead", buf, 8) == 0 )
      return;
    fprintf(out, "%s", buf);
    if (strncmp(buf, "score", 5) == 0)
      modeval = SCORE;
    else if (strncmp(buf, "music", 5) == 0)
      modeval = MUSIC;
    else if (strncmp(buf, "staff ", 6) == 0)
      modeval = STAFF;
    else {
      p = buf;
      while (*p == ' ' || *p == '\t')
	p++;
      if (strncmp(p, "staffs=", 6) == 0)
	donestaffs = 1;
      else if (strncmp(p, "time=", 5) == 0)
	extractTsig(p);			// get a default time sig
    }
  }
  fprintf(out, "// NO \"#headend\" STATEMENT?!\n");
}

void dosecondaries(FILE *fp) {
  event *e, *lp = NULL;
  char buf[80];

  while (fgets(buf, 80, fp) != NULL) {
    if (strncmp(buf, "#secend", 7) == 0 ||
	strncmp(buf, "#endsec", 7) == 0 )
      return;
    e = alloc_event();
    e->pitch = (int)strdup(buf);
    e->next = NULL;
    if (lp == NULL)
      secondaries = e;
    else
      lp->next = e;
    lp = e;
  }
}

void dobeamstyle(FILE *fp) {
  int j;
  char buf[80], *p;
  int n, d;
  
  while (fgets(buf, 80, fp) != '\0') {
    if (strncmp("#beamend", buf, 8) == 0 ||
	strncmp("#endbeam", buf, 8) == 0 )
      return;
    else {
      p = buf;
      while (*p == ' ' || *p == '\t')
	p++;
      n=0;
      d=0;
      while (isdigit(*p))
	n = 10 * n + *p++ - '0';
      if (*p != '/') {
	fprintf(stderr, "mtm: bad time signature in #beamstyle directive\n");
	continue;
      }
      p++;
      while (*p == ' ' || *p == '\t')
	p++;
      while (isdigit(*p))
	d = 10 * d + *p++ - '0';
      if (*p != '=') {
	fprintf(stderr, "mtm: bad time signature in #beamstyle directive\n");
	continue;
      }
      p++;
      bstyles[bwhre].bnum = n;
      bstyles[bwhre].bdnom = d;
      j = 0;
      for (j=0; j<79 && *p!='\n'; j++)
	bstyles[bwhre].style[j] = *p++;
      bstyles[bwhre].style[j] = 0;
      if (cflag == 0)
	fprintf(out, "// beamstyle %d/%d = %s\n", bstyles[bwhre].bnum,
		bstyles[bwhre].bdnom, bstyles[bwhre].style);
      bwhre++;
    }
  }
  fprintf(out, "// NO \"#beamend\" STATEMENT?!\n");
}

void dotranslate(char *buf) {
  int trk = 0, k;
  char *p, *q;

  p = buf+10;	// point at start of parameters
  while (*p == ' ' || *p == '\t')
    p++;
  if (!isdigit(*p)) {
    fprintf(stderr, "bad syntax in #translate directive - ignored\n");
    return;
  }
  while (isdigit(*p))
    trk = trk * 10 + *p++ - '0';
  if (trk >= MAXTRK) {
    fprintf(stderr, "track number too big in #translate directive - ignored\n");
    return;
  }
  while (*p == ' ' || *p == '\t')
    p++;
  if (*p != '=') {
    fprintf(stderr, "bad syntax in #translate directive - ignored\n");
    return;
  }
  p++;
  while (*p == ' ' || *p == '\t')
    p++;
  q = p + strlen(p) - 1;
  if (*q == '\n')
    q--;
  while (q > p && (*q == ' ' || *q == '\t'))
    q--;
  q[1] = 0;	// delete trailing whitespace
  k = strlen(p);
  if (k <= 0 || k > 62) {
    fprintf(stderr, "translation too long in #translate directive\n");
    return;
  }
  tlate[tlwhre].trk = trk;
  q = strchr(p, '/');		// are two maps provided
  if (q != NULL) {
    *q++ = 0;			// split into two, point q at second half
  }
  tlate[tlwhre].map1 = strdup(p);
  if (q != NULL)
    tlate[tlwhre].map2 = strdup(q);
  else
    tlate[tlwhre].map2 = tlate[tlwhre].map1;
  if (cflag == 0)
    fprintf(out, "// Translate track %d to %s.\n", trk, p);
  tlwhre++;
  return;
}

void dokey(char *buf) {
  int j, i;
  char ii[3], oo[3];

  for (i = 0; i < 3; i++) {
    ii[i] = '\0';
    oo[i] = '\0';
  }
  for (j = 4; ((buf[j] == ' ') || (buf[j] == '\t')); j++)
    ;
  
  ii[0] = buf[j];
  j++;
  if (buf[j] == '=') {
    j++;
    oo[0] = buf[j];
  }
  else {
    ii[1] = buf[j];
    j += 2;
    oo[0] = buf[j];
  }
  j++;
  if(buf[j] != '\n')
    oo[1] = buf[j];
  j = hkey[atoi(ii)] = atoi(oo);
  if (cflag == 0)
    fprintf(out, "// Key sig for track %d set to %d\n", atoi(ii), j);
  return;
}

static char *fmt_off(float v) {
  static char buff[16];
  char *p;
  
  sprintf(buff, "%f", v+1.0);
  p = buff + strlen(buff) - 1;
  while (*p == '0')
    p--;
  if (*p == '.')
    p--;
  p[1] = 0;
  return buff;
}

static void mode(int v) {
  if (modeval == v)
    return;
  modeval = v;
  switch (v) {
  case SCORE:
    fprintf(out, "score\n");
    break;
  case MUSIC:
    fprintf(out, "music\n");
    break;
  }
}

void usage(char *pgmname) {
  fprintf(stderr, "usage: %s [options] inputfile\n", pgmname);
  fprintf(stderr, "options currently available are:\n");
  fprintf(stderr, "\t-o filename = specify output file, '-' means stdout\n");
  fprintf(stderr, "\t-a quantum = specify quantisation factor\n");
  fprintf(stderr, "\t-s stretch = multiply note lengths before  quantising (default=0=no stretch)\n");
  fprintf(stderr, "\t\tnote that 'stretch' is an integer, with implied decimal point\n");
  fprintf(stderr, "\t\tso that -s100 is *1.00, -s125 adds a quarter, and so on.\n");
  fprintf(stderr, "\t-n = don't add measure number comments\n");
  fprintf(stderr, "\t-k num = specify a default key in range -7..7\n");
  fprintf(stderr, "\t-v = print version number and exit\n");
  exit(1);
}

static void doclef(char *p) {
  int trk, clef;
  p += 5;
  while (*p == ' ' || *p == '\t')
    p++;
  trk = 0;
  if (!isdigit(*p)) {
    fprintf(stderr, "missing track number in #clef directive - ignored\n");
    return;
  }
  while (isdigit(*p))
    trk = 10*trk + *p++ - '0';
  while (*p == ' ' || *p == '\t')
    p++;
  if (*p != '=') {
    fprintf(stderr, "syntax error in #clef directive - ignored\n");
    return;
  }
  p++;
  while (*p == ' ' || *p == '\t')
    p++;
  if (strncmp(p, "treble", 6) == 0)
    clef = 1;
  else if (strncmp(p, "bass", 4) == 0)
    clef = 2;
  else {
    fprintf(stderr, "syntax error in #clef directive - ignored\n");
    return;
  }
  if (trk < MAXTRK) {
    tracks[trk].clef = clef;
  }
  else
    fprintf(stderr, "track number(%d) in #clef directive is too big\n", trk);
}

static void inputParams(char *ifile) {
  FILE *fp;
  char buf[80], *p;

  fp = fopen(ifile, "r");
  if (fp == NULL) {
    ifile = ".midtomup";
    fp = fopen(ifile, "r");
  }
  if (fp != NULL) {
    while (fgets(buf, 80, fp) != '\0') {
      p = buf;
      while (*p == ' ' || *p == '\t')
	p++;
      if (*p == '\n')		// ignore null lines
	continue;
      if (p[0] != '#' && p[0] != '/') {
	fprintf(stderr, "Parse error in %s: %s\n", ifile, buf);
	break;
      }
      if (p[0] == '/' && p[1] == '/') {
	fprintf(out, "%s", buf);
	continue;
      }
      if (strncmp(p, "#header", 7) == 0)
	doheader(fp);
      else if (strncmp(p, "#secondaries", 12) == 0)
	dosecondaries(fp);
      else if (strncmp(p, "#beamstyle", 10) == 0)
	dobeamstyle(fp);
      else if (strncmp(p, "#translate", 10) == 0)
	dotranslate(p);
      else if (strncmp(p, "#key", 4) == 0)
	dokey(p);
      else if (strncmp(p, "#clef", 5) == 0)
	doclef(p);
      else
	fprintf(stderr, "Unknown rc command in %s: %s\n", ifile, buf);
    }
    fclose(fp);
  }
}

static int qcomp(const void *a, const void *b) {
  event *ap = *(event **)a;
  event *bp = *(event **)b;
  if (ap->start < bp->start)
    return -1;
  else if (ap->start > bp->start)
    return 1;
  /* obvious cases sorted, what remains is a->start == b->start.
     dispose of PATCH and TEXT items first */
  if (ap->type != NOTE)
    return -1;
  if (bp->type != NOTE)
    return 1;
  /* both are notes, return shortest before longest */
  else if (ap->end < bp->end)
    return -1;
  else if (ap->end > bp->end)
    return 1;
  else
    /* both same length, return lowest pitch first */
    if (ap->pitch < bp->pitch)
      return -1;
  else if (ap->pitch > bp->pitch)
    return 1;
  else
    return 0;		/* don't care - two identical notes ???? */
}

static void sorttracks(track *tracks) {
  int i, j, len;
  event **list, *e;
  
  for (i=0; i<MAXTRK; i++) {	// i scans through the given tracks
    if (tracks[i].notect > 0) {
      len = tracks[i].ct;
      list = (event **)malloc((len+1) * sizeof(event *));
      e = tracks[i].notes;
      for (j=0; j<len; j++) {
	list[j] = e;
	e = e->next;
      }
      qsort(list, len, sizeof(event*), qcomp);
      tracks[i].notes = list[0];
      for (j=0; j<len-1; j++) {
	list[j]->next = list[j+1];
	list[j+1]->prev = list[j];
      }
      list[j]->next = NULL;
      list[0]->prev = NULL;
      free(list);
      tracks[i].p = tracks[i].notes;
    }
  }
}

static char *maptrack(int i, int flag) {
  /* see if a track mapping has been defined */
  int j;	
  static char buff[32];

  for (j=0; j<tlwhre; j++)
    if (i == tlate[j].trk) {
      if (flag == 0)
	return tlate[j].map1;
      else
	return tlate[j].map2;
    }
  sprintf(buff, "%d", i);
  return buff;
}

static void resetvels(void) {
  velct = 0;
}

static void output_vels(int trk) {
  int i;

  for (i=0; i<velct; i++) {
    if (velvel[i] != vel[trk]) {
      float offset = (float)(velt[i] - h_ticks) / (float)division;

      if (velfudge == 0 && suppMidi == 0) {
	fprintf(out, "\nmidi %s: %s \"onvelocity=%d\";",
		maptrack(trk,1), fmt_off(offset), velvel[i]);
	}
      vel[trk] = velvel[i];
    }
  }
}

static int plen(int len) {
  // add note length to output file (if different from previous note).
  // If the given length is greater than one of the recognised lengths
  // return the excess (to be added later as a tie).
  float f;

  if (len == prevlen) {
    if (triplet > 0)
      triplet++;
    return 0;		/* same as previous note(s) - omit */
  }
  if (len == division)
    fprintf(out, "4");
  else if (len == (1.5 * division))
    fprintf(out, "4.");
  else if (len == (1.75 * division))
    fprintf(out, "4..");
  else if (len == division * 2)
    fprintf(out, "2");
  else if (len == division * 3)
    fprintf(out, "2.");
  else if (len == division * 3.5)
    fprintf(out, "2..");
  else if (len == division * 4)
    fprintf(out, "1");
  else if (len == (division/2))
    fprintf(out, "8");
  else if (len == (.75 * division))
    fprintf(out, "8.");
  else if (len == (.875 * division))
    fprintf(out, "8..");
  else if (len == (division/4))
    fprintf(out, "16");
  else if (len == (division/4 + division/8))
    fprintf(out, "16.");
  else if (len == (division/4 + division/8 + division/16))
    fprintf(out, "16..");
  else if (len == (division/8))
    fprintf(out, "32");
  else if (len == (division/16))
    fprintf(out, "64");
  else if (len == ((division*2)/3)) {
    if (triplet == 0)
      fprintf(out, "{4");
    else
      fprintf(out, "4");
    triplet++;
  }
  else if (len == (division/3)) {
    if (triplet == 0)
      fprintf(out, "{8");
    else
      fprintf(out, "8");
    triplet++;
  }
  else if (len == (division/6)) {
    if (triplet == 0)
      fprintf(out, "{16");
    else
      fprintf(out, "16");
    triplet++;
  }
  else if (len == (division/12)) {
    if (triplet == 0)
      fprintf(out, "{32");
    else
      fprintf(out, "32");
    triplet++;
  }
  else {
    /*
     * Try to handle some things that *can't* be a single note..
     * things that will have to be a tie, simply because there's
     * no other way to notate that length.  (Like a halfnote tied
     * to an eigthnote.)
     */
    if (len > 4 * division) {
      fprintf(out, "1");
      prevlen = 4 * division;
      return len - 4 * division;
    }
    if (len > 3 * division) {
      fprintf(out, "2.");
      prevlen = 3 * division;
      return len - 3 *division;
    }
    if (len > 2 * division) {
      fprintf(out, "2");
      prevlen = 2 * division;
      return len - 2 * division;
    }
    if (len > division + division/2) {
      fprintf(out, "4.");
      prevlen = (3 * division) / 2;
      return len - division - division/2;
    }
    /*
     * Some weird stuff to find counts like 2.5
     */
    if (len > division) {
      fprintf(out, "4");
      prevlen = division;
      return len - division;
    }
    else {
      if (len > division/2) {
	fprintf(out, "8");
	prevlen = division / 2;
	return(len - division/2);
      }
      if (len > division/4) {
	fprintf(out, "16");
	prevlen = division / 4;
	return len - division / 4;
      }
      if (len > division/8) {
	fprintf(out, "32");
	prevlen = division / 8;
	return len - division / 8;
      }
      if (division/16 > 0 && len > division/16) {
	fprintf(out, "64");
	prevlen = division / 16;
	return len - division / 16;
      }
      if (division/32 > 0 && len > division/32) {
	fprintf(out, "128");
	prevlen = division / 32;
	return len - division / 32;
      }
      if (division/64 > 0 && len > division/64) {
	fprintf(out, "256");
	prevlen = division / 64;
	return len - division / 64;
      }
      fprintf(out, "(BADLEN=%d,%d: quarters=%f)", len, division,
	      f=(float)len/(float)division);
    }
  }
  prevlen = len;
  return 0;
}

static void dorest(int start, int trck, int len) {
  int i, off, min = shortest;

  if (len > 0) {
    /* start is # of ticks from bar start */
    while (len > 0) {
      i = start / min;
      off = min;
      if (i == 0) {			/* any len OK */
	while (off * 2 <= len)	/* double while less than length */
	  off *= 2;
      }
      else
	while ((i & 1) == 0 && (off * 2) <= len) {
	  i = i >> 1;
	  off *= 2;
	}
      /* off is now largest factor of start, or == len */
      if (off != 0) {
	plen(off);
	len -= off;
	start += off;
	fprintf(out, "r;");
      }
    }
  }
  else
    fprintf(out, "mr;");
}

static void fixvel(long time, int vel) {
  int i;
  for (i=0; i<velct; i++) {
    if (velt[i] == time) {
      /* multiple notes at same time */
      if (velvel[i] < vel)
	velvel[i] = vel;
      return;
    }
  }
  if (velct >= MAXVEL) {
    fprintf(stderr, "mtm: velocity table overflow\n");
    exit(1);
  }
  if (velct == 0 || velvel[velct-1] != vel) {
    velt[velct] = time;
    velvel[velct++] = vel;
  }
}

void poctave(int trck, int nval) {
  /* for defoct=4 */
  if (key[trck] < -5)
    nval++;
  else if (key[trck] > 6)
    nval--;
  nval /= 12;
  if (tracks[trck].clef == 2)
    nval++;
  switch (nval) {
  case 0:
    fprintf(out, "-----");
    break;
  case 1:
    fprintf(out, "----");
    break;
  case 2:
    fprintf(out, "---");
    break;
  case 3:
    fprintf(out, "--");
    break;
  case 4:
    fprintf(out, "-");
    break;
  case 5:
    break;
  case 6:
    fprintf(out, "+");
    break;
  case 7:
    fprintf(out, "++");
    break;
  case 8:
    fprintf(out, "+++");
    break;
  case 9:
    fprintf(out, "++++");
    break;
  case 10:
    fprintf(out, "+++++");
    break;
  }
}

int goctave(int trck, int nval) {
  /* for defoct=4 */
  if (key[trck] < -5)
    nval++;
  else if (key[trck] > 6)
    nval--;
  nval /= 12;
  if (tracks[trck].clef == 2)
    nval++;	// handle defoct=3
  return nval;
}

static void donote(int trck, event *np, int num, int len) {
  // output a chord, with num notes, length=len, np->first event
  //  remaining events should be in sequence following...
  event *e;
  int i;
  char pitches[128];	// pitches used in this chord
  if (len == 0)
    return;		// ignore zero-length notes!
  do {
    int left = plen(len);

    len = len - left;
    bzero(pitches, 128);	// set no notes set
    for (e=np, i=0; i<num; i++, e=e->next) {
      if (e->start >= e->end) {
	e->start += len;
	continue;
      }
      fixvel(e->start, e->vel);
      e->start += len;
      if (pitches[e->pitch] == 0) {
	pitches[e->pitch] = 1;
	pnote(e->pitch, trck, e->tied);
	e->tied = 1;
	accwhre++;
	if (accwhre >= TABSIZE) {
	  fprintf(stderr, "oops!\n");
	  exit(1);
	}
	accnotes[accwhre] = e->pitch;
	poctave(trck, e->pitch);
	if (left != 0 || e->start < e->end ) {
	  fprintf(out, "~");
	  if (triplet == 3) {
	    fprintf(out, "}3");
	    triplet = 0;
	  }
	  tied[trck] = 1;
	}
      }
    }
    fprintf(out, ";");
    len = left;
  } while (len != 0);
  if (triplet == 3) {
    fprintf(out, "}3;");
    triplet = 0;
  }
}

static void do_notes(int i, track *ct, track *mct) {
  int j;
  event *np = ct->p;

  resetvels();
  if (np != NULL && np->type != END) {
    for (j = 0; j < MAXTRK; j++) {
      accnotes[j] = -1;
      sccnotes[j][0] = -1;
      sccnotes[j][1] = -1;
      sccnotes[j][2] = -1;
    }
    accwhre = -1;
    sccwhre = -1;
    ticks = h_ticks;
    /*
     * Do the requested track translations.
     */
    fprintf(out, "\n%s:", maptrack(i, 0));
    /* Is there a note in the measure? */
    if (np->start >= ticks && np->start < endticks) {
      do {
	if (np == NULL || np->type == END) {	/* end of list */
	  /* no more notes left, see if padding needed */
	  if (ticks < endticks) {
	    dorest(ticks-h_ticks, i, endticks - ticks);
	    ticks = endticks;
	  }
	}
	else if (np->type == NOTE) {
	  /* No, does a note start on current tick? */
	  if (np->start != ticks) {
	    /* Nope, so output a rest */
	    if (np->start >= endticks) {
	      dorest(ticks-h_ticks, i, endticks-ticks);
	      ticks = endticks;
	      break;
	    }
	    else {
	      dorest(ticks-h_ticks, i, np->start-ticks);
	      ticks = np->start;
	    }
	  }
	  /* having "rested" up to start of note... */
	  if (ticks < endticks) {
	    event *nxp = np->next;
	    int num = 1;
	    long fin, len;

	    fin = np->end;	// end of first note here
	    // see how many notes start at same time
	    while (nxp != NULL && np->start == nxp->start) {
	      if (nxp->end > ticks && nxp->end < fin)
		fin = nxp->end;
	      nxp = nxp->next;
	      num++;
	    }
	      /* fin now earliest finish time */
	    if (nxp != NULL && nxp->start < fin)
	      fin = nxp->start;
	    /* Note is tied to a note not in this measure */
	    if (fin > endticks)
	      fin = endticks;
	    len = np->end - np->start;	// length remaining
	    donote(i, np, num, fin-ticks);	// output 'num' notes
	    ticks = fin;
	    // np = nxp;
	  }
	  /* Next note in track, skip any zero of negative-length */
	  while (np->type == NOTE && np->start >= np->end)
	    np = np->next;
	}
	else {	/* PATCH or TEXT */
	  if (np->type != END)
	    np = np->next;
	}
      } while (ticks < endticks && np->start < endticks && np->type != END);
      if (ticks < endticks) {
	dorest(ticks-h_ticks, i, endticks - ticks);
	ticks = endticks;
      }
    }
    else {
      dorest(0, i, MEASURE);
    }
  }
  tracks[i].p = np;
  output_vels(i);
}

FILE *efopen(char *name) {
  FILE *f;
  const char *errmess;
  char fname[128];

  if ((f=fopen(name, "rb")) == NULL) {
    sprintf(fname, "%s.mid", name);
    if ((f = fopen(fname, "rb")) == NULL) {
      errmess = strerror(errno);
      (void) fprintf(stderr,"Cannot open '%s'!\n",name);
      (void) fprintf(stderr,"Reason: %s\n",errmess);
      exit(1);
    }
  }
  return(f);
}

static void do_args(int argc, char **argv) {
  char ofile[80], *outfname=NULL, *p;
  char *errmess;
  int c;

  while ((c = getopt(argc, argv, "a:cdhk:mno:s:vV:wz")) != EOF)
    switch(c) {
    case 'd':
      debug = 1;
      break;
    case 'k':
      dfltkey = atoi(optarg);
      if (dfltkey > 7 || dfltkey < -7) {
	fprintf(stderr, "mtm: default key setting out of range -7..7\n");
	exit(1);
      }
      midikey[0].when = 0;
      midikey[0].nkey = dfltkey;	// stick a key change at start
      mwhre++;
      break;
    case 'a':
      ffactor = atoi(optarg) / 4;
      break;
    case 'c':
      cflag = 1;		// inhibit most comments
      break;
    case 'o':
      outfname = optarg;
      break;
    case 's':
      stretch = atoi(optarg);
      break;
    case 'm':
      suppMidi = 0;
      break;
    case 'n':
      pmeasure = 0;
      break;
    case 'v':
      fprintf(stderr, "%s\n", version);
      exit(0);
      break;
    case 'V':
      velfudge = atoi(optarg);
      break;
    case 'w':
      warn = 0;
      break;
    case 'z':
      zflag = 1;
      break;
    default:
      fprintf(stderr, "-%c not understood.\n", c);
    case 'h':
      usage(argv[0]);
      exit(1);
    }
  if (optind >= argc) {
    usage(argv[0]);
    exit(1);
  }
  strcpy(ifile, argv[optind]);
  p = strrchr(ifile, '.');
  if (p != NULL)
    *p = 0;
  if (outfname == NULL) {
    strcpy(ofile, ifile);
    strcat(ofile, ".mup");
    outfname = ofile;
  }
  strcat(ifile, ".mtm");
  inf = efopen(argv[optind]);
  if (strcmp(outfname, "-") == 0)
    out = stdout;
  else {
    out = fopen(outfname, "w");
    if (out == NULL) {
      errmess = strerror(errno);
      fprintf(stderr, "mtm: unable to open output file: %s\n", outfname);
      fprintf(stderr, "Reason: %s\n", errmess);
      exit(1);
    }
  }
  bytecount = -1;
}

void compact(void) {
  // scan down the tracks from 1-31, discarding any with notect=0
  // shunt non-zero tracks down to fill gaps, set trackno=new upper limit
  int i, k;
  for (i=1,k=1; i<MAXTRK; i++) {
    if (tracks[i].notect > 0) {
      if (i > k) {
	tracks[k]  = tracks[i];
	mtrks[k] = mtrks[i];
      }
      k++;
    }
  }
  trackno = k;
}
    
void dobar(void) {
  fprintf(out, "\nbar\n");
}

static void formatDate(void) {
  // ctime does: Wed Nov 11 13:56:21 2005 (alter to suit local conditions)
  long tt = time(NULL);
  char *t = ctime(&tt);
  fputs(t, out);
}
  
int main(int argc, char **argv) {
  event *e;
  int i, j;

  for (i = 0; i < MAXTRK; i++) {
    tracks[i].ct = 0;		// no events in track
    tracks[i].notect = 0;	// no notes in track
    tracks[i].notes = NULL;	// set pointer to NULL
    tracks[i].p     = NULL;	// set work pointer to NULL;
    tracks[i].minp = 128;	// initialise min and max pitch
    tracks[i].maxp = -1;	// ready for testing
    mtrks[i].ct = 0;		// no events in auxiliary track
    mtrks[i].notes = NULL;	// set pointer to NULL
    mtrks[i].p = NULL;		// ditto for work pointer
  }
  twhre = 0;
  bwhre = 0;
  tlwhre = 0;
  stretch = 0;
  ffactor=0;		// reasonable default, can be overridden
  do_args(argc, argv);
  mfread();		// scan the midifile, building note events
  fclose(inf);
  /*
   * Done reading the midi file.  Now output the notes.
   *
   * First output a 'standard' Mup header message
   */
  fprintf(out, "//!Mup-Arkkra\n");
  fprintf(out, "// generated by mtm version %s on ", VERSION);
  formatDate();
  /*
   * Next, read the control file if present ...
   */
  inputParams(ifile);
  sorttracks(tracks);
  sorttracks(mtrks);
  compact();		// discard any tracks with notect=0 (apart from 0)
  
  /* Init some vars */
  for (i=1; i<trackno; i++) {
    tr[i] = 0;
    vel[i] = -1;
    if (tracks[i].clef == 0) {	// has it been defined in control file?
      if (tracks[i].maxp < 65)
	tracks[i].clef = 2;	// looks like bass clef to me
      else
	tracks[i].clef = 1;	// safer to use Treble clef
    }
  }
  measure = 1;
  ticks = 0;
  maxtimesig = twhre;
  twhre = 0;
  maxkeysig = mwhre;		/* note number of key sigs read */
  mwhre = 0;			/* reset to first one */
  if (maxkeysig == 0) {		/* no key sigs in file? */
    for (i=0; i<MAXTRK; i++)
      setkeysig(i, dfltkey);	/* set all tracks to C Major */
  }
  h_ticks = 0;
  triplet = 0;
  /* If no time sig specified, insert one, set to 4/4 */
  if (maxtimesig <= 0) {
    tsig[0].num = 4;
    tsig[0].dnom = 4;
    maxtimesig=1;
  }
  mode(SCORE);
  if (!donestaffs)
    fprintf(out, "\tstaffs=%d\n", trackno-1);
  e = secondaries;
  while (e) {
    fprintf(out, "%s", (char *)e->pitch);
    e = e->next;
  }
  if (secondaries != NULL)
    modeval = -1;		/* secondaries may have left us in limbo? */
  for (i=1; i<trackno; i++) {
    if (tracks[i].clef == 2) {
      fprintf(out, "staff %s\n\tclef=bass\n", maptrack(i, 1));
      modeval = STAFF;
    }
  }
  /* Main loop to output notes */
  do {
    while (twhre < maxtimesig && tsig[twhre].when <= h_ticks) {
      mode(SCORE);
      fprintf(out, "\ttime=%d/%d\n", tsig[twhre].num, tsig[twhre].dnom);
      for (i = 0; i < 50; i++) {
	if (bstyles[i].bnum == tsig[twhre].num &&
	    bstyles[i].bdnom == tsig[twhre].dnom) {
	  fprintf(out, "\tbeamstyle=%s\n", bstyles[i].style);
	  break;
	}
      }
      twhre++;
    }
    while (mwhre < maxkeysig && midikey[mwhre].when <= h_ticks) {
      mode(SCORE);
      if (midikey[mwhre].nkey < 0)
	fprintf(out, "\tkey=%d&\n", -midikey[mwhre].nkey);
      else
	fprintf(out, "\tkey=%d#\n", midikey[mwhre].nkey);
      for (i = 1; i < trackno; i++) {
	j = midikey[mwhre].nkey + hkey[i];
	if (tracks[i].ct > 0 && hkey[i] != 0) {
	  /*
	   * Print the staff number, remembering translations..
	   */
	  fprintf(out, "staff %s\n", maptrack(i, 1));
	  if (j <= 0)
	    fprintf(out, "\tkey=%d&\n", -j);
	  else
	    fprintf(out, "\tkey=%d#\n", j);
	}
	key[i] = j;
	setkeysig(i, j);
      }
      mwhre++;
    }
    mode(MUSIC);
    if (pmeasure)
      fprintf(out, "\n// bar %d", measure);
    measure++;
    /*
     * Sorry, only quarters and eighths and halfs can be in the denominator
     */
    if (tsig[twhre - 1].dnom == 4)
      endticks = h_ticks + (division * tsig[twhre - 1].num);
    else if (tsig[twhre - 1].dnom == 8)
      endticks = h_ticks + (division/2 * tsig[twhre - 1].num);
    else
      endticks = h_ticks + (division * 2 * tsig[twhre - 1].num);
    /*
     * We will deal with one measure at a time, one track at a time.
     */
    for (i=0; i<trackno; i++)
      resetaccs(i);		/* reset accidentals */
    for (ct=tracks, mct=mtrks, i=0; i<trackno; i++,ct++,mct++)	{
      event *np = mct->p;
      prevlen = -1;
      if (np != NULL && np->type != END) {
	if (suppMidi == 0) {
	  /* first, scan bar for midi commands */
	  while (np && np->start < endticks) {
	    float offset = (float)(np->start - h_ticks) / (float)division;
	    
	    if (np->type == TEMPO)
	      fprintf(out, "\nmidi all: %s \"tempo=%ld\";",
		      fmt_off(offset), np->pitch);
	    else if (i > 0) {
	      if (np->type == PATCH)
		fprintf(out, "\nmidi %s: %s \"program=%ld\";",
			maptrack(i,1), fmt_off(offset), np->pitch);
	      else if (np->type == TEXT )
		fprintf(out, "\nmidi %s: %s \"text=%s\";",
			maptrack(i,1), fmt_off(offset), (char *)(np->vel));
	      else if (np->type == PITCH)
		fprintf(out, "\nmidi %s: %s \"chanpressure=%ld\";",
			maptrack(i,1), fmt_off(offset), np->pitch);
	      else if (np->type == PARAM)
		fprintf(out, "\nmidi %s: %s \"param=%ld,%ld\";",
			maptrack(i,1), fmt_off(offset), np->pitch, np->vel);
	      else if (np->type == BEND)
		fprintf(out, "\nmidi %s: %s \"text=bend=%ld\";",
			maptrack(i,1), fmt_off(offset), (np->pitch*128)+np->vel);
	    }
	    np = np->next;
	  }
	  mct->p = np;
	}
      }
      /* reset to start of bar */
      if (i > 0)
	do_notes(i, ct, mct);
    }
    dobar();
    h_ticks = endticks;
  } while (h_ticks < lasttick);
  if (discards > 0) {
    fprintf(stderr, "mtm: %ld note(s) discarded as too short\n", discards);
  }
  exit(0);
}

// the remaining functions are all invoked from midifile.c during
// input of the midi file.

void mtm_error(char *s) {
  fprintf(stderr,"Error: %s\n",s);
}

void mtm_header(int fmt, int ntrks, int ldivision) {
  format = fmt;
  division = ldivision;
  if (ffactor == 0) {
    // no quantum specified, find a reasonable value
    // for this division. Note: some synths produce
    // really grotty values, just divide by 2 until
    // we get an odd number, then double that
    ffactor = division /2;
    while ((ffactor & 2) == 0)
      ffactor /= 2;
  }
  else if (ffactor < 2)
    ffactor = 0;
  else
    ffactor = division / ffactor;		/* derive quantum in ticks */
  shortest = division;
  while ((shortest & 1) == 0)
    shortest = shortest >> 1;
  /* shortest is now smallest possible note length */
  /*
    printf("Header format=%d ntrks=%d division=%d\n",format,ntrks,division);
    */
}

void mtm_trackstart(void) {
  /*
    printf("Track start\n");
    */
  ct = tracks + trackno;
  mct = mtrks + trackno;
  trackno++;
  ct->p = NULL;
  ct->notes = NULL;
  mct->p = NULL;
  mct->notes = NULL;
}

long adjust(long t) {
  if (ffactor == 0)
    return t;
  return ((t + (ffactor >> 1) - 1) / ffactor) * ffactor;
}

// this routine only provided for debugging, to be invoked by patches
// or manually from gdb
void checkCount(track *t) {
  int c=0;
  event *e = t->notes;
  while (e != NULL) {
    if (e->type == NOTE)
      c++;
    e = e->next;
  }
  if (c != t->notect) {
    printf("consistency check failure\n");
  }
}

static void add_to_trk(track *ct, event *e) {
  if (ct->p == NULL)
    ct->notes = e;
  else
    ct->p->next = e;
  e->prev = ct->p;
  ct->p = e;
  ct->ct++;
}

static void discard(track *ct, event *e) {
  // remove event from track (probably because length=0
  if (e->prev == NULL)
    ct->notes = e->next;
  else
    e->prev->next = e->next;
  if (e->next != NULL)
    e->next->prev = e->prev;
  if (ct->p == e)
    ct->p = e->prev;
  free_event(e);
  ct->notect--;
  ct->ct--;
  discards++;
}

static void finishTrack(track *ct, track *mct) {
  long t = adjust(Mf_currtime);
  event *e = alloc_event();

  e->start = t + division;
  e->next = NULL;
  e->type = END;
  add_to_trk(ct, e);
  
  e = alloc_event();
  e->start = t + division;
  e->next = NULL;
  e->type = END;
  add_to_trk(mct, e);
}

void mtm_trackend(void) {
  /* append dummy entry to mark end-of-track */
  int i;

  if (format == 0) {
    for (i=0; i<trackno; i++)
      finishTrack(tracks+i, mtrks+i);
  }
  else
    finishTrack(ct, mct);
}

static void prtime(void) {
  int t = Mf_currtime / division, f = Mf_currtime % division;
  
  printf("Time=%d.%d \n", t, f);
}

void mtm_noteon(int chan, int pitch, int vol) {
  long t = Mf_currtime, len;
  event *e;
  
  if (debug) {
    prtime();
    printf("Note on, chan=%d pitch=%d vol=%d\n",chan+1,pitch,vol);
  }
  if (format == 0) {
    ct = tracks+chan+1;		/* split notes into tracks based on channel */
    mct = mtrks+chan+1;
    if (trackno <= chan+1)
      trackno = chan+2;
  }
  if (vol == 0) {		/* note off message */
    if (ct->p == NULL) {
      if (warn) {
	prtime();
	printf("end note before first start note\n");
      }
      return;	// ignore end note event
    }
    for (e=ct->p; e; e=e->prev)
      if (e->type == NOTE && e->pitch == pitch && e->end < 0)
	break;
    if (e == NULL) {
      int octave = pitch / 12;
      int note = pitch - 12 * octave;
  
      if (warn) {
	prtime();
	printf("can't find matching start note, trk=%d, pitch=%s%d\n",
	       chan, nnam[note], octave);
      }
      return;			/* discard unmatched note */
    }
    while (e->start >= t)	/* extend zero-length notes to minimum */
      t += ffactor;
    if (stretch > 0) {
      len = t - e->start;
      len += (len * stretch) / 100;
      t = e->start + len;
    }
    e->end = adjust(t);
    if (lasttick < t)
      lasttick = t;
    if (e->end <= e->start) {
      // zero or negative length note ??
      if (zflag)
	discard(ct, e);
      else
	e->end = e->start + ffactor;
    }
  }
  else {
    e = alloc_event();
    e->start = adjust(t);
    e->vel = vol;
    e->pitch = pitch;
    e->type = NOTE;
    e->end = -1;		/* unterminated */
    add_to_trk(ct, e);
    ct->notect++;
    if (pitch < ct->minp)
      ct->minp = pitch;
    if (pitch > ct->maxp)
      ct->maxp = pitch;
  }
}

void mtm_noteoff(int chan, int pitch, int vol) {
  mtm_noteon(chan, pitch, 0);	/* map all real note-offs to noteon with vel=0 */
}

void mtm_pressure(int chan, int pitch, int press) {
  /*
    prtime();
    printf("Pressure, chan=%d pitch=%d press=%d\n",chan+1,pitch,press);
    */
}

void mtm_program(int chan, int program) {
  /*
    prtime();
    printf("Program, chan=%d program=%d\n",chan+1,program);
    */
  long t = adjust(Mf_currtime);
  event *e = alloc_event();

  e->start = t;
  e->type = PATCH;
  e->pitch = program;
  add_to_trk(mct, e);
}

void mtm_parameter(int chan, int control, int value) {
  /*
    prtime();
    printf("Parameter, chan=%d c1=%d c2=%d\n",chan+1,control,value);
    */
  long t = adjust(Mf_currtime);
  event *e = alloc_event();
  e->start = t;
  e->type = PARAM;
  e->pitch = control;
  e->vel = value;
  add_to_trk(mct, e);
}

void mtm_pitchbend(int chan, int msb, int lsb) {
  /*
    prtime();
    printf("Pitchbend, chan=%d msb=%d lsb=%d\n",chan+1,msb,lsb);
    */
  long t = adjust(Mf_currtime);
  event *e = alloc_event();
  e->start = t;
  e->type = BEND;
  e->pitch = msb;
  e->vel = lsb;
  add_to_trk(mct, e);
}

void mtm_chanpressure(int chan, int press) {
  long t = adjust(Mf_currtime);
  event *e = alloc_event();

  e->start = t;
  e->type = PITCH;
  e->pitch = press;
  add_to_trk(mct, e);
  /*
    prtime();
    printf("Channel pressure, chan=%d pressure=%d\n",chan+1,press);
    */
}

void mtm_sysex(int leng, char *mess) {
  /*
    prtime();
    printf("Sysex, leng=%d\n",leng);
    */
}

void mtm_metamisc(int type, int leng, char *mess) {
  /*
    prtime();
    printf("Meta event, unrecognized, type=0x%02x leng=%d\n",type,leng);
    */
}

void mtm_seqspecific(int type, int leng, char *mess) {
  /*
    prtime();
    printf("Meta event, sequencer-specific, type=0x%02x leng=%d\n",type,leng);
    */
}

void mtm_text(int type, int leng, char *mess) {
  static char *ttype[] = {
    NULL,
    "Text Event",		/* type=0x01 */
    "Copyright Notice",	/* type=0x02 */
    "Sequence/Track Name",
    "Instrument Name",	/* ...       */
    "Lyric",
    "Marker",
    "Cue Point",		/* type=0x07 */
    "Unrecognized"
  };
  int unrecognized = (sizeof(ttype)/sizeof(char *)) - 1;
  long t = adjust(Mf_currtime);
  event *e = alloc_event();

  if ( type < 1 || type > unrecognized )
    type = unrecognized;

  e->start = t;
  e->type = TEXT;
  e->pitch = type;
  mess[leng] = 0;		/* just in case! */
  e->vel = (int)strdup(mess);
  add_to_trk(mct, e);
}

void mtm_seqnum(int num) {
  /*
    prtime();
    printf("Meta event, sequence number = %d\n",num);
    */
}

void mtm_eot(void) {
  /*
    prtime();
    printf("Meta event, end of track\n");
    */
}

void mtm_keysig(int sf, int mi) {
  /*
    prtime();
    printf("Key signature, sharp/flats=%d  minor=%d\n",sf,mi);
    */
  if (mwhre >= MAXKEYSIG) {
    fprintf(stderr, "mtm: too many key signature changes\n");
    exit(1);
  }
  midikey[mwhre].when=Mf_currtime;
  /* the ? stuff is to compensate for compilers that don't sign
   * extend the sf */
  midikey[mwhre].nkey = (sf & 0x80) ? (sf | ~0xff) : sf;
  mwhre++;
}

void mtm_tempo(long ltempo) {
  event *e = alloc_event();
  long t = adjust(Mf_currtime);

  e->type = TEMPO;
  e->pitch = 60000000 / ltempo;
  e->start = t;
  add_to_trk(mct, e);
}

void mtm_timesig(int nn, int dd, int cc, int bb) {
  int denom = 1;
  if (ct != tracks)		// ignore timesigs on other than trk 0
    return;
  
  while ( dd-- > 0 )
    denom *= 2;
  /*
    prtime();
    printf("Time signature=%d/%d  MIDI-clocks/click=%d"
           "  32nd-notes/24-MIDI-clocks=%d\n",
	   nn,denom,cc,bb);
   */
  tsig[twhre].when=Mf_currtime;
  tsig[twhre].num=nn;
  tsig[twhre].dnom = denom;
  twhre++;
}

void mtm_smpte(int hr, int mn, int se, int fr, int ff) {
  /*
    prtime();
    printf("SMPTE, hour=%d minute=%d second=%d frame=%d fract-frame=%d\n",
           hr,mn,se,fr,ff);
   */
}

void mtm_arbitrary(int leng, char *mess)
{
  /*
    prtime();
    printf("Arbitrary bytes, leng=%d\n",leng);
    */
}

/* read and return one character from MIDI file. Return EOF on end of file
 * or read error. */
int mtm_getc(void) {
  int ch = fgetc(inf);
  bytecount++;
  debugbuff[bytecount & 15] = ch;
  return ch;
}

// this routine provided for debugging, invoked manually from gdb.
void dumpTrack(event *ep) {
  while (ep != NULL) {
    printf("event 0x%08lx: next=0x%08lx, type=%d, start=%ld, end=%ld\n",
	   (long)ep, (long)(ep->next), ep->type, ep->start, ep->end);
    ep = ep->next;
  }
}
